/*
 * Giuseppe Caulo - <giusc@softzone.it>
 * http://www.softzone.it
 */

/*
 * Based on original Code by HatemMostafa
 */

#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/stat.h>
#include <time.h>

#ifdef _WIN32
# define snprintf _snprintf
# define strcasecmp stricmp
# ifndef __STDC__
#  define stat _stat
#  define lstat _stat
#  define fstat _fstat
#  define S_IFMT   _S_IFMT
#  define S_IFDIR  _S_IFDIR
#  define S_IFCHR  _S_IFCHR
#  define S_IFREG  _S_IFREG
#  define S_IREAD  _S_IREAD
#  define S_IWRITE _S_IWRITE
#  define S_IEXEC  _S_IEXEC
# endif
# define S_ISREG(x)  ((x) & _S_IFREG)
# define S_ISDIR(x)  ((x) & _S_IFDIR)
#endif



#define GetBit(a, b)	((((BYTE*)a)[(b)>>3] >> ((b)&7)) & 1)
#define SetBit(a, b)	(((BYTE*)a)[(b)>>3] |= (1 << ((b)&7)))

class CHuffmanNode
{
public:
	CHuffmanNode() {	memset(this, 0, sizeof(CHuffmanNode));	}

	int nFrequency;	// must be first for byte align
	BYTE byAscii;
	DWORD dwCode;
	int nCodeLength;
	CHuffmanNode *pParent, *pLeft, *pRight;
};

int __cdecl frequencyCompare(const void *elem1, const void *elem2 )
{
	CHuffmanNode *pNodes[2] = {	(CHuffmanNode*)elem1, (CHuffmanNode*)elem2 };
	if(pNodes[0]->nFrequency == pNodes[1]->nFrequency)
		return 0;
	return pNodes[0]->nFrequency < pNodes[1]->nFrequency ? 1 : -1;
}

int __cdecl asciiCompare(const void *elem1, const void *elem2 )
{
	return ((CHuffmanNode*)elem1)->byAscii > ((CHuffmanNode*)elem2)->byAscii ? 1 : -1;
}

CHuffmanNode* PopNode(CHuffmanNode *pNodes[], int nIndex, bool bRight)
{
	CHuffmanNode* pNode = pNodes[nIndex];
	pNode->dwCode = bRight;
	pNode->nCodeLength = 1;
	return pNode;
}

void SetNodeCode(CHuffmanNode* pNode)
{
	CHuffmanNode* pParent = pNode->pParent;
	while(pParent && pParent->nCodeLength)
	{
		pNode->dwCode <<= 1;
		pNode->dwCode |= pParent->dwCode;
		pNode->nCodeLength++;
		pParent = pParent->pParent;
	}
}

int GetHuffmanTree(CHuffmanNode nodes[], bool bSetCodes = true)
{
	CHuffmanNode* pNodes[256], *pNode;
	// add used ascii to huffman queue
	int nNodeCount = 0;
	for(int nCount = 0; nCount < 256 && nodes[nCount].nFrequency; nCount++)
		pNodes[nNodeCount++] = &nodes[nCount];
	int nParentNode = nNodeCount, nBackNode = nNodeCount-1;
	while(nBackNode > 0)
	{
		// parent node
		pNode = &nodes[nParentNode++];
		// pop first child
		pNode->pLeft = PopNode(pNodes, nBackNode--, false);
		// pop second child
		pNode->pRight = PopNode(pNodes, nBackNode--, true);
		// adjust parent of the two poped nodes
		pNode->pLeft->pParent = pNode->pRight->pParent = pNode;
		// adjust parent frequency
		pNode->nFrequency = pNode->pLeft->nFrequency + pNode->pRight->nFrequency;
		// insert parent node depending on its frequency
		for(int i = nBackNode; i >= 0; i--)
			if(pNodes[i]->nFrequency >= pNode->nFrequency)
				break;
		memmove(pNodes+i+2, pNodes+i+1, (nBackNode-i)*sizeof(int));
		pNodes[i+1] = pNode;
		nBackNode++;
	}
	// set tree leaves nodes code
	if(bSetCodes)
		for(nCount = 0; nCount < nNodeCount; nCount++)
			SetNodeCode(&nodes[nCount]);

	return nNodeCount;
}

bool CompressHuffman(BYTE *pSrc, int nSrcLen, BYTE *&pDes, int &nDesLen)
{
	CHuffmanNode nodes[511];
	// initialize nodes ascii
	for(int nCount = 0; nCount < 256; nCount++)
		nodes[nCount].byAscii = nCount;
	// get ascii frequencies
	for(nCount = 0; nCount < nSrcLen; nCount++)
		nodes[pSrc[nCount]].nFrequency++;
	// sort ascii chars depending on frequency
	qsort(nodes, 256, sizeof(CHuffmanNode), frequencyCompare);
	// construct huffman tree
	int nNodeCount = GetHuffmanTree(nodes);
	// construct compressed buffer
	int nNodeSize = sizeof(DWORD)+sizeof(BYTE);
	nDesLen = nSrcLen+nNodeCount*nNodeSize;
	pDes = (BYTE*)malloc(nDesLen);
	BYTE *pDesPtr = pDes;
	memset(pDesPtr, 0, nDesLen);
	// save source buffer length at the first DWORD
	*(DWORD*)pDesPtr = nSrcLen;
	pDesPtr += sizeof(DWORD);
	// save huffman tree leaves count-1 (as it may be 256)
	*pDesPtr = nNodeCount-1;
	pDesPtr += sizeof(BYTE);
	// save huffman tree used leaves nodes
	for(nCount = 0; nCount < nNodeCount; nCount++)
	{	// the array sorted on frequency so used nodes come first
		memcpy(pDesPtr, &nodes[nCount], nNodeSize);
		pDesPtr += nNodeSize;
	}
	// sort nodes depending on ascii to can index nodes with its ascii value
	qsort(nodes, 256, sizeof(CHuffmanNode), asciiCompare);

	int nDesIndex = 0, nCodeLength, dwCode;
	// loop to write codes
	for(nCount = 0; nCount < nSrcLen; nCount++)
	{
		dwCode = nodes[pSrc[nCount]].dwCode;
		nCodeLength = nodes[pSrc[nCount]].nCodeLength;
		while(nCodeLength--)
		{
			if(dwCode&1)
				SetBit(pDesPtr, nDesIndex);
			dwCode >>= 1, nDesIndex++;
		}
	}
	// update destination length
	nDesLen = (pDesPtr-pDes)+(nDesIndex+7)/8;
	pDes = (BYTE*)realloc(pDes, nDesLen);

	return true;
}

bool DecompressHuffman(BYTE *pSrc, int nSrcLen, BYTE *&pDes, int &nDesLen)
{
	// copy destination final length
	nDesLen = *(DWORD*)pSrc;
	// allocate buffer for decompressed buffer
	pDes = (BYTE*)malloc(nDesLen+1);
	int nNodeCount = *(pSrc+sizeof(DWORD))+1;
	// initialize huffman nodes with frequency and ascii
	CHuffmanNode nodes[511], *pNode;
	int nNodeSize = sizeof(DWORD)+sizeof(BYTE), nSrcIndex = nNodeSize;
	for(int nCount = 0; nCount < nNodeCount; nCount++)
	{
		memcpy(&nodes[nCount], pSrc+nSrcIndex, nNodeSize);
		nSrcIndex += nNodeSize;
	}
	// construct huffman tree
	GetHuffmanTree(nodes, false);
	// get huffman tree root
	CHuffmanNode *pRoot = &nodes[0];
	while(pRoot->pParent)
		pRoot = pRoot->pParent;

	int nDesIndex = 0;
	nSrcIndex <<= 3;	// convert from bits to bytes
	while(nDesIndex < nDesLen)
	{
		pNode = pRoot;
		while(pNode->pLeft)	// if node has pLeft then it must has pRight
		{	// node not leaf
			pNode = GetBit(pSrc, nSrcIndex) ? pNode->pRight : pNode->pLeft;
			nSrcIndex++;
		}
		pDes[nDesIndex++] = pNode->byAscii;
	}

	return true;
}


void Compress(LPCSTR lpcsSrcFile, LPCSTR lpcsDesFile, bool bCompress)
{
	int nFile = _open(lpcsSrcFile, _O_RDONLY|_O_BINARY, 0);
	int nFileLength = _filelength(nFile);
	BYTE *pFileBuffer = new BYTE[nFileLength];
	_read(nFile, pFileBuffer, nFileLength);
	_close(nFile);

	// compress or decompress file buffer
	BYTE *pDes;
	int nDesLen;
	if(bCompress)
		CompressHuffman(pFileBuffer, nFileLength, pDes, nDesLen);
	else
		DecompressHuffman(pFileBuffer, nFileLength, pDes, nDesLen);
	delete pFileBuffer;

	// open destination buffer
	nFile = _open(lpcsDesFile, _O_CREAT|_O_BINARY|_O_WRONLY, _S_IREAD|_S_IWRITE);
	// write file buffer
	_write(nFile, pDes, nDesLen);
	_close(nFile);

	free(pDes);
}


int Usage( char *ProgramName )
{
    printf("Usage: %s [-d] [-s] <inputfile> <outputfile>\n", ProgramName );
    printf("-d     Decompress\n");
    printf("-s     Silent, do not show Banner\n\n");
    return -1;
}


int main(int argc, char *argv[])
{
	bool bCompress = 1;
	bool bShowBanner = 1;
	clock_t start, finish;
	unsigned long units;

	if( argc > 2 ) {
		for(int i = 1; i < argc; i++ ) {
			if( argv[i][0] != '/' && argv[i][0] != '-' ) {
				if ( (i != argc-1) && (i != argc-2) ) {
					printf( argv[i] );
					return Usage( argv[0] );
				}
				continue;
			}
			if( argv[i][1] == 'd' || argv[i][1] == 'D' ) {
				bCompress = false;
			} else if ( argv[i][1] == 's' || argv[i][1] == 'S' ) {
				bShowBanner = false;
			}
			else return Usage( argv[0] );
		}
	} else {
		return Usage( argv[0] );
	}
	if (bShowBanner) {
		printf("\nHuffMan v1.0 - Compress/Decompress Using Huffman method\n" );
		printf("Copyright (C) 2005 Giuseppe Caulo\n");
		printf("Softzone - www.softzone.it\n\n");
	}
    start = clock();
    printf("\n%s -> %s\n\n", argv[argc-2], argv[argc - 1]);
	Compress(argv[argc-2], argv[argc-1], bCompress);
	finish = clock();
	units = (unsigned long)finish - start;
	printf("took %ld time units (%f seconds)\n", units, (double)units / (double)CLOCKS_PER_SEC);
}